"""The core can provide information through this module to the user."""
# dem/cli/core_cb.py

from dem.cli.console import stdout
import typer, typing
from rich.progress import Progress, TaskID, TextColumn, BarColumn, TaskProgressColumn
from rich.status import Status

def get_value_by_key_if_exist(dictionary: dict, keys: list[str]) -> str | TaskID | None:
    """ Returns with the value based on the key if it exists, None if not.
    
    Args:
        dictionary -- the dictionary to get the value from
        keys -- list of keys that addresses a nested value
    """
    for key in keys:
        try:
            dictionary = dictionary[key]
        except KeyError:
            return None

    return dictionary

def msg_cb(*args, msg: str, rule: bool = False) -> None:
    """ Generic callback function to present information for the user. 
    
    Args:
        args -- (not used) the function acts as a class method, so it gets the caller objects as 
                input
        msg -- the message to print
        rule -- indicate the start of a new section
    """
    if rule is True:
        stdout.rule(msg)
    else:
        stdout.print(msg)

def user_confirm_cb(*args, **kwargs) -> None:
    """ Callback function to get confirmation from the user
    
    Args:
        args -- (not used) the function acts as a class method, so it gets the caller objects as 
                input
        kwargs -- msg: message to print
                  user_confirm: the action the user needs to confirm
    """
    if "msg" in kwargs:
        stdout.print(kwargs["msg"])

    if "user_confirm" in kwargs:
        typer.confirm(kwargs["user_confirm"], abort=True)

def update_progress_bar(tasks: dict, id: str, progress: Progress, item: dict, status: str) -> None:
    """ Update the progress bar of the image pull.
    
    Args:
        tasks -- dictionary of tasks (task represent a row in the Progress table, which contains the
                 progress bar)
        id -- layer id
        progress -- contains the progress bars
        item -- current item from the generator
        status -- which process's status the progress bar shows
    """
    task = get_value_by_key_if_exist(tasks, [id])

    if task is None:
        task = progress.add_task(str(id), id=id)
        tasks[id] = task

    current = get_value_by_key_if_exist(item, ["progressDetail", "current"])
    total = get_value_by_key_if_exist(item, ["progressDetail", "total"])

    if current and total:
        progress.update(task, description=str(status), total=float(total), completed=float(current))

def process_generator_item(item: dict, tasks: dict, progress: Progress) -> None:
    """ Process an item from the generator provided by the pull command.
    
    Args:
        item-- current item from the generator
        tasks -- dictionary of tasks
        progress -- contains the progress bars
        """
    status = get_value_by_key_if_exist(item, ["status"])
    id = get_value_by_key_if_exist(item, ["id"])

    if status:
        if id:
            if get_value_by_key_if_exist(item, ["progressDetail"]):
                update_progress_bar(tasks, id, progress, item, status)
            else:
                progress.console.print(str(id) + ": " + str(status))
        else:
            progress.console.print(str(status))

def pull_progress_cb(*args, generator: typing.Generator) -> None:
    """ Process the pull command's generator

    Args:
        args -- (not used) the function acts as a class method, so it gets the caller objects as 
                input
        generator -- the pull command's generator
    """
    with Progress(TextColumn("[progress.layer_id]{task.fields[id]}"), 
                  TextColumn("[progress.description]{task.description}"),
                  BarColumn(), TaskProgressColumn(), console=stdout) as progress:
        tasks = {}

        for item in generator:
            process_generator_item(item, tasks, progress)

status = Status("")
def status_start_cb(*args, status_msg: str) -> None:
    status.update(status=status_msg)
    status.start()

def status_stop_cb(*args) -> None:
    status.stop()



"""This module represents the Development Environments."""
# dem/core/dev_env_setup.py

from dem.core.exceptions import InvalidDevEnvJson
from dem.core.properties import __supported_dev_env_major_version__
from dem.core.tool_images import ToolImages
from dem.core.data_management import LocalDevEnvJSON, OrgDevEnvJSON
from dem.core.container_engine import ContainerEngine

class DevEnv:
    """ A Development Environment."""
    supported_tool_types = ( 
        "build system",
        "toolchain",
        "debugger",
        "deployer",
        "test framework",
    )

    def _check_tool_type_support(self, descriptor: dict) -> None:
        """ Check that the Dev Env doesn't contain an unsupported tool type.
        
            Private function that gets called on instantiation.
            Args:
                descriptor -- the description of the Development Environment from the dev_env.json 
                              file
        """
        for tool in descriptor["tools"]:
            if tool["type"] not in self.supported_tool_types:
                raise InvalidDevEnvJson("The following tool type is not supported: " + tool["type"])

    def __init__(self, descriptor: dict) -> None:
        """ Init the DevEnv class.
        
            Args:
                descriptor -- the description of the Development Environment from the dev_env.json 
                              file
        """
        self._check_tool_type_support(descriptor)
        self.name = descriptor["name"]
        self.tools = descriptor["tools"]

    def check_image_availability(self, tool_images: ToolImages, update_tool_images: bool = False) -> list:
        """ Checks the tool image's availability.
        
            Updates the "image_status" key for the tool dictionary.
            Returns with the statuses of the Dev Env tool images.

            Args:
                tool_images -- the images the Dev Envs can access
                update_tool_images -- update the list of available tool images
        """
        if update_tool_images == True:
            tool_images.update()
        image_statuses = []
        for tool in self.tools:
            tool_image_name = tool["image_name"] + ':' + tool["image_version"]
            image_status = tool_images.NOT_AVAILABLE
            if tool_image_name in tool_images.elements:
                image_status = tool_images.elements[tool_image_name]
            image_statuses.append(image_status)
            tool["image_status"] = image_status

        return image_statuses

class DevEnvSetup:
    """Representation of the development setup:
        - The available tool images.
        - The available Development Environments.

    Class attributes:
        _tool_images -- the available tool images
        _container_engine -- the container engine
    """
    _tool_images = None
    _container_engine = None
    pull_progress_cb = None

    def _dev_env_json_version_check(self) -> None:
        """Check that the dev_env.json or dev_evn_org.json file supported.

        The dev_env.json or dev_env_org.json version is stored as string in the X.Y format.
        Raises an InvalidDevEnvJson exception, if the version is not supported.
        """
        dev_env_json_major_version = int(self.version.split('.', 1)[0])
        if dev_env_json_major_version != __supported_dev_env_major_version__:
            raise InvalidDevEnvJson("The dev_env.json version v1.0 is not supported.")

    def __init__(self, dev_env_json_deserialized: dict) -> None:
        """Create a list for the Development Environments.

        Args:
            dev_env_json_deserialized -- a deserialized representation of the dev_env.json or 
                                         dev_env_org.json file
        """
        self.version = dev_env_json_deserialized["version"]
        self._dev_env_json_version_check()
        self.dev_envs = []

    @classmethod
    @property
    def tool_images(cls) -> ToolImages:
        """ Only instantiate the _tool_images when first accessed, because it is time consuming.

        The decorators are needed so the tool_images can act as a class-level property, ensuring 
        only a single ToolImage class intance exists for every DevEnvSetup() instance. The class 
        variable can be accessed through a getter, so the ToolImage() gets instantiated only at the 
        first access.

        Args:
            cls - the class object
        """
        if cls._tool_images is None:
            cls._tool_images = ToolImages(cls.container_engine)
        return cls._tool_images

    @classmethod
    @property
    def container_engine(cls) -> ContainerEngine:
        """ The container engine.

        The decorators are needed so the container_engine can act as a class-level property, 
        ensuring only a single ContainerEngine class intance exists for every DevEnvSetup() instance. 
        The class variable can be accessed through a getter, so the ContainerEngine() gets 
        instantiated only at the first access.

        Args:
            cls - the class object
        """
        if cls._container_engine is None:
            cls._container_engine = ContainerEngine()

        return cls._container_engine

    def get_deserialized(self) -> dict:
        """ Create the deserialized json. """
        dev_env_json_deserialized = {}
        dev_env_json_deserialized["version"] = self.version
        dev_env_descriptors = []
        for dev_env in self.dev_envs:
            dev_env_descriptor = {}
            dev_env_descriptor["name"] = dev_env.name
            dev_env_descriptor["tools"] = dev_env.tools
            dev_env_descriptors.append(dev_env_descriptor)
        dev_env_json_deserialized["development_environments"] = dev_env_descriptors
        return dev_env_json_deserialized

    def get_dev_env_by_name(self, dev_env_name: str) -> ("DevEnvOrg | DevEnvLocal | None"):
        """Get the Development Environment by name.
        
        Args:
            dev_env_name -- name of the Development Environment to get
        Returns with the instance representing the Development Environment. If the Development 
        Environment doesn't exist in the setup, the function returns with None.
        """
        for dev_env in self.dev_envs:
            if dev_env.name == dev_env_name:
                return dev_env

class DevEnvLocal(DevEnv):
    """ Local Development Environment """
    def __init__(self, descriptor: dict | None = None, dev_env_org: "DevEnvOrg | None" = None):
        """Init a local DevEnv class

        The class can be initialized either based on the Dev Env descriptor from the dev_env.json 
        file or on an already existing DevEnvOrg object.
        Args:
            dev_env_json_deserialized -- a deserialized representation of the dev_env.json file
            dev_env_org -- the DevEnvOrg object to make a copy from (note: forward reference)
        """
        if descriptor is not None:
            super().__init__(descriptor)
        else:
            self.name = dev_env_org.name
            self.tools = dev_env_org.tools

class DevEnvLocalSetup(DevEnvSetup):
    """ The local development setup.

    Class attributes:
        json -- deserialized json representing the local setup
        invalid_json_cb -- the user can be noted through this callback when something is wrong with
                           the dev_env.json file
        msg_cb -- a generic callback to send messages for the user
        pull_progress_cb -- a generator is provided for this callback to note the user of the 
                            current status of the pull process
    """
    json = LocalDevEnvJSON()
    invalid_json_cb = None
    msg_cb = None
    pull_progress_cb = None
    status_start_cb = None
    status_stop_cb = None

    def __init__(self):
        """ Store the local Development Environments.

        Extends the DevEnvSetup super class by populating the list of Development Environments with 
        DevEnvLocal objects.
        """
        if self.invalid_json_cb:
            self.json.set_invalid_json_callback(self.invalid_json_cb)

        if self.pull_progress_cb:
            self.container_engine.set_pull_progress_cb(self.pull_progress_cb)

        if (self.status_start_cb is not None) and (self.status_stop_cb is not None):
            ToolImages.status_start_cb = self.status_start_cb
            ToolImages.status_stop_cb = self.status_stop_cb

        super().__init__(self.json.read())

        for dev_env_descriptor in self.json.deserialized["development_environments"]:
            self.dev_envs.append(DevEnvLocal(descriptor=dev_env_descriptor))

    def update_json(self):
        """ Writes the deserialized json to the dev_env.json file."""
        self.json.write(self.get_deserialized())

    def pull_images(self, tools: list):
        """ Pull images that are only present in the registry.
        
        Args:
            tools -- the tool images to pull (with any status, this function will filter the 
                     registry only ones)
        """

        # Remove the duplications and the locally already available tools
        unique_tool_images_to_pull = []
        for tool in tools:
            image_to_pull = tool["image_name" ] + ':' + tool["image_version"]
            if (image_to_pull not in unique_tool_images_to_pull) and \
                (tool["image_status"] == ToolImages.REGISTRY_ONLY):
                unique_tool_images_to_pull.append(image_to_pull)

        for tool_image in unique_tool_images_to_pull:
            if self.msg_cb is not None:
                self.msg_cb(msg="\n")
                self.msg_cb(msg="Pulling image " + tool_image, rule=True)

            self.container_engine.pull(tool_image)

class DevEnvOrg(DevEnv):
    """A Development Environment available for the organization."""
    def get_local_instance(self, dev_env_setup_local: DevEnvLocalSetup) -> (DevEnvLocal | None):
        """Check if this Development Enviroment already installed locally.

        Args:
            dev_env_setup_local -- the locally installed Development Environments
        Returns the local Dev Env object if available. None if not yet installed.
        """
        for dev_env_local in dev_env_setup_local.dev_envs:
            if self.name == dev_env_local.name:
                return dev_env_local

class DevEnvOrgSetup(DevEnvSetup):
    """ The organization's development setup. The user can install Dev Envs listed in this class.

    Class attributes:
        json -- deserialized json representing the organization's setup 
    """
    json = OrgDevEnvJSON()

    def __init__(self) -> None:
        """Store the Development Environments available for the organization.

        Extends the DevEnvSetup super class by populating the list of Development Environments with 
        DevEnvOrg objects.
        """
        super().__init__(self.json.read())

        for dev_env_descriptor in self.json.deserialized["development_environments"]:
            self.dev_envs.append(DevEnvOrg(descriptor=dev_env_descriptor))



"""Direct access to the registry over HTTP. 
Use the container engine when possible for accessing the registry. 
"""
# dem/core/registry.py

import requests
from dem.core import container_engine as container_engine
from dem.core.exceptions import RegistryError

def list_repos(container_engine_obj: container_engine.ContainerEngine, 
               status_start_cb = None, status_stop_cb = None) -> list[str]:
    registry = "axemsolutions"
    images = []
    
    if status_start_cb is not None:
        status_start_cb(status_msg="Loading image information from the registry...")

    for image in container_engine_obj.search(registry):
        url = f"https://registry.hub.docker.com/v2/repositories/{image}/tags/"

        response = requests.get(url)

        if response.status_code == requests.codes.ok:
            for result in response.json()["results"]:
                images.append(image + ":" + result["name"])
        else:
            raise RegistryError("Error in communication with the registry. Failed to retrieve tags. Response status code: ",
                                response.status_code)

    if status_stop_cb is not None:
        status_stop_cb()

    return images


"""Docker container engine management."""
# dem/core/container_engine.py

from typing import Callable
from types import MethodType
import docker

class ContainerEngine():
    def __init__(self) -> None:
        """Operations on the Docker Container Engine."""
        self._docker_client = docker.from_env()
        self._docker_api_client = docker.APIClient(base_url="unix:///var/run/docker.sock", 
                                                   version="auto")
        self._pull_progress_cb = None

    def get_local_tool_images(self) -> list[str]:
        """Get local tool images.
        
        Return with the list of the locally avialable tool image names.
        """
        local_image_tags = []

        for image in self._docker_client.images.list():
            for tag in image.tags:
                if tag:
                    local_image_tags.append(tag)

        return local_image_tags

    def pull(self, repository: str) -> None:
        """Pull a repository from the axemsolutions registry.
        
        Args:
            repository -- repository to pull"""
        if self._pull_progress_cb:
            resp = self._docker_api_client.pull(repository, stream=True, decode=True)
            self._pull_progress_cb(generator=resp)
        else:
            self._docker_api_client.pull(repository)

    def remove(self, image: str) -> None:
        """Remove a tool image.
        
        Args: 
            image -- the tool image to remove"""
        self._docker_client.images.remove(image)

    def search(self, registry: str) -> None:
        """Search repository in the axemsolutions registry.
        
        Args:
            registry -- registry to search"""
        local_registryimagelist = []

        for repositories in self._docker_client.images.search(registry):        
            local_registryimagelist.append(repositories['name'])
        return local_registryimagelist

    def set_pull_progress_cb(self, pull_progress_callback: Callable) -> None:
        self._pull_progress_cb = MethodType(pull_progress_callback, self)



"""dev_env.json handling."""
# dem/core/data_management.py

from pathlib import PurePath
from typing import Callable
import os, types
import json, requests

_empty_dev_env_json = """{
    "version": "0.1",
    "org_name": "axem",
    "registry": "registry-1.docker.io",
    "development_environments": []
}
"""

class LocalDevEnvJSON():
    """ Serialize and deserialize the dev_env.json file."""
    _path = PurePath(os.path.expanduser('~') + "/.config/axem/dev_env.json")
    _directory = PurePath(os.path.expanduser('~') + "/.config/axem")

    def _create_empty_dev_env_json(self) -> dict:
        """ If the dev_env.json doesn't exist, then create an empty one."""
        is_path_exist = os.path.exists(self._directory)
        if not is_path_exist:
            os.makedirs(self._directory)

        json_file = open(self._path, "w")
        json_file.write(_empty_dev_env_json)
        json_file.close()

        return json.loads(_empty_dev_env_json)

    @staticmethod
    def _invalid_json_callback(*args, **kwargs) -> None:
        pass

    def __init__(self) -> None:
        """ Init the class with an empty dict for the deserialized dev_env.json file. 
            Later this variable can be used to access the deserialized data. 
        """
        self.deserialized = {}

    def read(self) -> dict:
        """ Read the deserialized dev_env.json."""
        try:
            json_file = open(self._path, "r")
        except FileNotFoundError:
            self.deserialized = self._create_empty_dev_env_json()
        else:
            try:
                self.deserialized = json.load(json_file)
            except json.decoder.JSONDecodeError:
                self._invalid_json_callback(msg="[red]Error: invalid json format.[/]", 
                                            user_confirm="Restore the original json file?")
                self.deserialized = self._create_empty_dev_env_json()
            else:
                json_file.close()
        
        return self.deserialized

    def write(self, deserialized: dict) -> None:
        """ Write modified deserialized data to the dev_env.json file.
            
            Args:
                deserialized -- the modified deserialized data
        """
        self.deserialized = deserialized
        json_file = open(self._path, "w")
        json.dump(deserialized, json_file, indent=4)
        json_file.close()

    def set_invalid_json_callback(self, invalid_json_callback: Callable):
        self._invalid_json_callback = types.MethodType(invalid_json_callback, self)

    def set_callback(self, callback_func: Callable):
        self._callback = types.MethodType(callback_func, self)

class OrgDevEnvJSON():
    """ Deserialize the dev_env_org.json file."""
    def __init__(self) -> None:
        """ Init the class with an empty placeholder for the deserialized dev_env_org.json file. 
            Later this variable can be used to access the deserialized data. 
        """
        self.deserialized = None

    def read(self) -> dict:
        """ Read the deserialized dev_env_org.json from the axemsolutions domain."""
        response = requests.get("https://axemsolutions.io/dem/dev_env_org.json")
        self.deserialized = response.json()
        return self.deserialized


"""This module provides the CLI."""
# dem/cli/main.py

import typer, importlib.metadata
from dem import __command__, __app_name__
from dem.cli.command import info_cmd, list_cmd, pull_cmd, create_cmd, modify_cmd, delete_cmd, rename_cmd, clone_cmd
from dem.cli.console import stdout

typer_cli = typer.Typer()

@typer_cli.command()
def list(local: bool = typer.Option(False, help="Scope is the local host."),
         all: bool = typer.Option(False, help="Scope is the organization."),
         env: bool = typer.Option(False, help="List the environments."),
         tool: bool = typer.Option(False, help="List the tool images.")) -> None:
    """
    List the Development Environments available locally or for the organization.
    
    The following option combinations are suppported:

        --local --env -> List the local Development Environments.

        --all --env -> List the organization's Development Environments.

        --local --tool -> List the local tool images.

        --all --tool -> List the tool images available in the axemsolutions registry.
    """
    list_cmd.execute(local, all, env, tool)

@typer_cli.command()
def info(dev_env_name: str = typer.Argument(...,
                                            help="Name of the Development Environment to get info about.")) -> None:
    """
    Get information about the specified Development Environment.
    """
    info_cmd.execute(dev_env_name)

@typer_cli.command()
def pull(dev_env_name: str = typer.Argument(..., 
                                            help="Name of the Development Environment to install.")) -> None:
    """
    Pull all the required tool images from the registry and install the Development Environment
    locally.
    """
    pull_cmd.execute(dev_env_name)

@typer_cli.command()
def clone(dev_env_name: str = typer.Argument(...,help="Name of the Development Environment to clone."),
           new_dev_env_name: str = typer.Argument(...,help="Name of the New Development Environment.")) -> None:
    """
    Clone existing Development Environment locally.
    """
    clone_cmd.execute(dev_env_name,new_dev_env_name)

@typer_cli.command()
def create(dev_env_name: str = typer.Argument(..., 
                                              help="Name of the Development Environment to create."),) -> None:
    """
    Create a new Development Environment.
    """
    create_cmd.execute(dev_env_name)

@typer_cli.command()
def rename(dev_env_name: str = typer.Argument(...,help="Name of the Development Environment to rename."),
           new_dev_env_name: str = typer.Argument(...,help="The new name.")) -> None:
    """
    Rename the Development Environment.
    """
    rename_cmd.execute(dev_env_name,new_dev_env_name)

@typer_cli.command()
def modify(dev_env_name: str = typer.Argument(..., 
                                              help="Name of the Development Environment to modify.")) -> None:
    """
    Modify the tool types and required tool images for an existing Development Environment.
    """
    modify_cmd.execute(dev_env_name)

@typer_cli.command()
def delete(dev_env_name: str = typer.Argument(..., 
                                              help="Name of the Development Environment to delete.")) -> None:
    """
    Delete the Development Environment from the local setup. If a tool image is not required
    anymore by any of the available local Development Environments, the DEM asks the user whether
    they want to delete that image or not.
    """
    delete_cmd.execute(dev_env_name)

def _version_callback(value: bool) -> None:
    if value:
        try: 
            version = importlib.metadata.version(__app_name__)
        except importlib.metadata.PackageNotFoundError:
            stdout.print("[yellow]Install DEM to get the version number.[/]")
        else: 
            stdout.print("[cyan]" + __app_name__ + " v" + version + "[/]")

        raise typer.Exit()

@typer_cli.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the dem version.",
        callback=_version_callback,
        is_eager=True,
    )) -> None:
    """
    Development Environment Manager (dem)
    
    Manage your containerized Development Environments with ease.



    ❗ Always put the input text into double quotation marks (""), if it contains whitespaces.

    """
    return


"""create CLI command implementation."""
# dem/cli/command/create_cmd.py

import typer
from dem.core.dev_env_setup import DevEnv, DevEnvLocal, DevEnvLocalSetup
from dem.core.tool_images import ToolImages
from dem.cli.console import stdout, stderr
from dem.cli.tui.panel.tool_type_selector import ToolTypeSelectorPanel
from dem.cli.tui.panel.tool_image_selector import ToolImageSelectorPanel

tool_image_statuses = {
    ToolImages.LOCAL_ONLY: "local",
    ToolImages.REGISTRY_ONLY: "registry",
    ToolImages.LOCAL_AND_REGISTRY: "local and registry"
}

def get_tool_image_list(tool_images: ToolImages) -> list[list[str]]:
    return [[name, tool_image_statuses[status]] for name, status in tool_images.elements.items()]

def handle_tool_type_selector_panel(tool_type_selector_panel: ToolTypeSelectorPanel, 
                                    dev_env_name: str) -> list[str]:
    tool_type_selector_panel.tool_type_menu.set_title("What kind of tools would you like to include in [cyan]" + dev_env_name + "[/]?")

    tool_type_selector_panel.wait_for_user()

    if "cancel" in tool_type_selector_panel.cancel_next_menu.get_selection():
        raise(typer.Abort())

    tool_type_selector_panel.cancel_next_menu.is_selected = False

    return tool_type_selector_panel.tool_type_menu.get_selected_tool_types()

def handle_tool_image_selector_panel(tool_image_selector_panel: ToolImageSelectorPanel, 
                                     tool_type:str) -> str | None:
    tool_image_selector_panel.tool_image_menu.set_title("Select tool image for: [yellow]" + tool_type + "[/]")
    tool_image_selector_panel.wait_for_user()

    if tool_image_selector_panel.back_menu.is_selected is True:
        # Reset the back menu selection
        tool_image_selector_panel.back_menu.is_selected = False
        return None
    else:
        tool_image_selector_panel.tool_image_menu.is_selected = False
        return tool_image_selector_panel.tool_image_menu.get_selected_tool_image()

def get_dev_env_descriptor_from_user(dev_env_name: str, tool_image_list: list[list[str]]) -> dict:
    current_panel = ToolTypeSelectorPanel(list(DevEnv.supported_tool_types))
    panel_list = [current_panel]

    tool_index = 0
    panel_index = 0
    tool_selection = {}
    while current_panel is not None:
        if isinstance(current_panel, ToolTypeSelectorPanel):
            selected_tool_types = handle_tool_type_selector_panel(current_panel, dev_env_name)

            # Remove the not selected tool type from the tool_selection.
            for tool_type in list(tool_selection.keys()):
                if tool_type not in selected_tool_types:
                    del tool_selection[tool_type]

            if len(panel_list) > 1:
                current_panel = panel_list[1]
                current_panel.dev_env_status.reset_table(selected_tool_types)
            else:
                current_panel = ToolImageSelectorPanel(tool_image_list, selected_tool_types)
                panel_list.append(current_panel)

            tool_index = 0
            panel_index = 1
        else:
            selected_tool_image = handle_tool_image_selector_panel(current_panel, selected_tool_types[tool_index])

            if selected_tool_image is None:
                tool_selection[selected_tool_types[tool_index]] = "<not selected>"

                panel_index -= 1
                current_panel = panel_list[panel_index]
                
                if tool_index != 0:
                    tool_index -= 1
            else:
                tool_selection[selected_tool_types[tool_index]] = selected_tool_image

                tool_index += 1
                
                if tool_index == len(selected_tool_types):
                    break

                panel_index += 1
                if len(panel_list) > panel_index:
                    current_panel = panel_list[panel_index]
                else:
                    current_panel = ToolImageSelectorPanel(tool_image_list, selected_tool_types)
                    panel_list.append(current_panel)

                current_panel.dev_env_status.reset_table(selected_tool_types)

            if isinstance(current_panel, ToolImageSelectorPanel):
                current_panel.dev_env_status.set_tool_image(tool_selection)

    dev_env_descriptor = {
        "name": dev_env_name,
        "tools": []
    }

    for tool_type, tool_image in tool_selection.items():
        tool_descriptor = {
            "type": tool_type,
            "image_name": tool_image.split(":")[0],
            "image_version": tool_image.split(":")[1]
        }
        dev_env_descriptor["tools"].append(tool_descriptor)

    return dev_env_descriptor

def overwrite_existing_dev_env(original_dev_env: DevEnvLocal, new_dev_env_descriptor: dict) -> None:
    original_dev_env.tools = new_dev_env_descriptor["tools"]

def create_new_dev_env(dev_env_local_setup: DevEnvLocalSetup, new_dev_env_descriptor: dict) -> DevEnvLocal:
    new_dev_env = DevEnvLocal(new_dev_env_descriptor)
    dev_env_local_setup.dev_envs.append(new_dev_env)

    return new_dev_env

def create_dev_env(dev_env_local_setup: DevEnvLocalSetup, dev_env_name: str) -> DevEnvLocal:
    dev_env_original = dev_env_local_setup.get_dev_env_by_name(dev_env_name)
    if dev_env_original is not None:
        typer.confirm("The input name is already used by a Development Environment. Overwrite it?", 
                      abort=True)

    tool_image_list = get_tool_image_list(dev_env_local_setup.tool_images)
    new_dev_env_descriptor = get_dev_env_descriptor_from_user(dev_env_name, tool_image_list)
    
    if dev_env_original is not None:
        overwrite_existing_dev_env(dev_env_original, new_dev_env_descriptor)
        new_dev_env = dev_env_original
    else:
        new_dev_env = create_new_dev_env(dev_env_local_setup, new_dev_env_descriptor)

    new_dev_env.check_image_availability(dev_env_local_setup.tool_images)
    dev_env_local_setup.pull_images(new_dev_env.tools)

    return new_dev_env

def execute(dev_env_name: str) -> None:
    dev_env_local_setup = DevEnvLocalSetup()
    dev_env = create_dev_env(dev_env_local_setup, dev_env_name)

    # Validate the Dev Env creation
    image_statuses = dev_env.check_image_availability(dev_env_local_setup.tool_images, 
                                                      update_tool_images=True)

    if (ToolImages.NOT_AVAILABLE in image_statuses) or (ToolImages.REGISTRY_ONLY in image_statuses):
        stderr.print("The installation failed.")
    else:
        stdout.print("The [yellow]" + dev_env.name + "[/] Development Environment is ready!")
        dev_env_local_setup.update_json()


"""list CLI command implementation."""
# dem/cli/list_cmd.py

from dem.core import container_engine, dev_env_setup, registry
from dem.core.tool_images import ToolImages
from dem.cli.console import stdout, stderr
from rich.table import Table

(
    DEV_ENV_ORG_NOT_IN_REGISTRY,
    DEV_ENV_ORG_INSTALLED_LOCALLY,
    DEV_ENV_ORG_REAINSTALL,
    DEV_ENV_ORG_READY,
) = range(4)

(
    DEV_ENV_LOCAL_NOT_AVAILABLE,
    DEV_ENV_LOCAL_REINSTALL,
    DEV_ENV_LOCAL_INSTALLED,
) = range(3)

dev_env_org_status_messages = {
    DEV_ENV_ORG_NOT_IN_REGISTRY: "[red]Error: Required image is not available in the registry![/]",
    DEV_ENV_ORG_INSTALLED_LOCALLY: "Installed locally.",
    DEV_ENV_ORG_REAINSTALL: "Incomplete local install. Reinstall needed.",
    DEV_ENV_ORG_READY: "Ready to be installed.",
}

dev_env_local_status_messages = {
    DEV_ENV_LOCAL_NOT_AVAILABLE: "[red]Error: Required image is not available![/]",
    DEV_ENV_LOCAL_REINSTALL: "Incopmlete local install. Reinstall needed.",
    DEV_ENV_LOCAL_INSTALLED: "Installed.",
}

def is_dev_env_org_installed_locally(dev_env_org: dev_env_setup.DevEnvOrg) -> bool:
    dev_env_local_setup_obj = dev_env_setup.DevEnvLocalSetup()
    return dev_env_org.get_local_instance(dev_env_local_setup_obj) is not None

def get_dev_env_status(dev_env: (dev_env_setup.DevEnvLocal | dev_env_setup.DevEnvOrg), 
                       tool_images: ToolImages) -> str:
    image_statuses = dev_env.check_image_availability(tool_images)
    dev_env_status = ""
    if isinstance(dev_env, dev_env_setup.DevEnvOrg):
        if (ToolImages.NOT_AVAILABLE in image_statuses) or (ToolImages.LOCAL_ONLY in image_statuses):
            dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_NOT_IN_REGISTRY]
        elif (image_statuses.count(ToolImages.LOCAL_AND_REGISTRY) == len(image_statuses)) and \
                (is_dev_env_org_installed_locally(dev_env) == True):
            dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_INSTALLED_LOCALLY]
        else:
            if (is_dev_env_org_installed_locally(dev_env) == True):
                dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_REAINSTALL]
            else:
                dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_READY]
    else:
        if (ToolImages.NOT_AVAILABLE in image_statuses):
            dev_env_status = dev_env_local_status_messages[DEV_ENV_LOCAL_NOT_AVAILABLE]
        elif (ToolImages.REGISTRY_ONLY in image_statuses):
            dev_env_status = dev_env_local_status_messages[DEV_ENV_LOCAL_REINSTALL]
        else:
            dev_env_status = dev_env_local_status_messages[DEV_ENV_LOCAL_INSTALLED]
    return dev_env_status

def list_dev_envs(local: bool, org: bool)-> None:
    dev_env_setup_obj = None
    if ((local == True) and (org == False)):
        dev_env_setup_obj = dev_env_setup.DevEnvLocalSetup()
        if not dev_env_setup_obj.dev_envs:
            stdout.print("[yellow]No installed Development Environments.[/]")
            return
    elif((local == False) and (org == True)):
        dev_env_setup_obj = dev_env_setup.DevEnvOrgSetup()
        if not dev_env_setup_obj.dev_envs:
            stdout.print("[yellow]No Development Environment in your organization.[/]")
            return
    else:
        stderr.print("[red]Error: Invalid options.[/]")
        return

    table = Table()
    table.add_column("Development Environment")
    table.add_column("Status")
    for dev_env in dev_env_setup_obj.dev_envs:
        table.add_row(dev_env.name, get_dev_env_status(dev_env, dev_env_setup_obj.tool_images))

    stdout.print(table)

def list_tool_images(local: bool, org: bool) -> None:
    container_engine_obj = container_engine.ContainerEngine()
    if (local == True) and (org == False):        
        local_images = container_engine_obj.get_local_tool_images()

        table = Table()
        table.add_column("Repository")
        for local_image in local_images:
            table.add_row(local_image)
        stdout.print(table)
    elif (local == False) and (org == True):
        registry_images = registry.list_repos(container_engine_obj)

        table = Table()
        table.add_column("Repository")
        for registry_image in registry_images:
            table.add_row(registry_image)
        stdout.print(table)

def execute(local: bool, org: bool, env: bool, tool: bool) -> None:
    if ((local == True) or (org == True)) and (env == True) and (tool == False):
        list_dev_envs(local, org)
    elif ((local == True) or (org == True)) and (env == False) and (tool == True):
        list_tool_images(local, org)
    else:
        stderr.print(\
"""Usage: dem list [OPTIONS]
Try 'dem list --help' for help.

Error: You need to set the scope and what to list!""")


"""delete CLI command implementation."""
# dem/cli/command/delete_cmd.py

from dem.core.dev_env_setup import DevEnvLocalSetup, DevEnvLocal
from dem.cli.console import stderr, stdout
import typer
import docker.errors

def try_to_delete_tool_image(tool_image: str, dev_env_local_setup: DevEnvLocalSetup) -> None:
    stdout.print("\nThe tool image [bold]" + tool_image + "[/bold] is not required by any Development Environment anymore.")
    if typer.confirm("Would you like to remove it?"):
        try:
            dev_env_local_setup.container_engine.remove(tool_image)
        except docker.errors.ImageNotFound:
            stdout.print("[yellow]" + tool_image + " doesn't exist. Unable to remove it.[/]\n")
        except docker.errors.APIError:
            stderr.print("[red]Error: " + tool_image + " is used by a container. Unable to remove it.[/]\n")
        else:
            stdout.print("[green]Successfully removed![/]\n")

def remove_unused_tool_images(deleted_dev_env: DevEnvLocal, dev_env_local_setup: DevEnvLocalSetup) -> None:
    all_required_tool_images = set()
    for dev_env in dev_env_local_setup.dev_envs:
        for tool in dev_env.tools:
            all_required_tool_images.add(tool["image_name"] + ":" + tool["image_version"])

    deleted_dev_env_tool_images = set()
    for tool in deleted_dev_env.tools:
        deleted_dev_env_tool_images.add(tool["image_name"] + ":" + tool["image_version"])

    for tool_image in deleted_dev_env_tool_images:
        if tool_image not in all_required_tool_images:
            try_to_delete_tool_image(tool_image, dev_env_local_setup)

def execute(dev_env_name: str) -> None:
    dev_env_local_setup = DevEnvLocalSetup()
    dev_env_to_delete = dev_env_local_setup.get_dev_env_by_name(dev_env_name)

    if dev_env_to_delete is None:
        stderr.print("[red]Error: The [bold]" + dev_env_name + "[/bold] Development Environment doesn't exist.")
    else:
        stdout.print("Deleting the Development Environment...")
        dev_env_local_setup.dev_envs.remove(dev_env_to_delete)
        dev_env_local_setup.update_json()

        remove_unused_tool_images(dev_env_to_delete, dev_env_local_setup)
        stdout.print("[green]Successfully deleted the " + dev_env_name + "![/]")


"""Local and registry tool images."""
# dem/core/tool_images.py

import dem.core.registry as registry
from dem.cli.console import stdout
from dem.core.exceptions import RegistryError
from dem.core.container_engine import ContainerEngine

class ToolImages():
    """The available tool images for a Dev Env.
    
    Provides the list of tool images available locally and in the registry.
    """
    (
        LOCAL_ONLY,
        REGISTRY_ONLY,
        LOCAL_AND_REGISTRY,
        NOT_AVAILABLE,
    ) = range(4)

    status_start_cb = None
    status_stop_cb = None

    def __init__(self, container_engine: ContainerEngine) -> None:
        """Init the ToolImages class with the up to date image statuses."""
        self.elements = {}
        self.container_egine = container_engine

        self.update()

    def update(self) -> None:
        """Update the image statuses based on the local and registry tool availability."""
        self.elements = {}

        for local_image in self.container_egine.get_local_tool_images():
            self.elements[local_image] = self.LOCAL_ONLY

        try:
            registry_images = registry.list_repos(self.container_egine, 
                                                  self.status_start_cb, self.status_stop_cb)
        except RegistryError as e:
            stdout.print("[red]" + str(e) + "[/]")
            stdout.print("[red]Using local tool images only![/]")
            registry_images = []

        for registry_image in registry_images:
            if registry_image in self.elements:
                self.elements[registry_image] = self.LOCAL_AND_REGISTRY
            else:
                self.elements[registry_image] = self.REGISTRY_ONLY


"""Entry point for dem."""
# dem/__main__.py

from dem import __command__
from dem.cli.console import stderr, stdout
from dem.core.exceptions import RegistryError
from dem.core.dev_env_setup import DevEnvLocalSetup
import dem.cli.main, dem.cli.core_cb
import docker.errors
import types

def main():
    """ Entry point for the CLI application"""

    # Set callback for core modules.
    DevEnvLocalSetup.invalid_json_cb = dem.cli.core_cb.user_confirm_cb
    DevEnvLocalSetup.msg_cb = dem.cli.core_cb.msg_cb
    DevEnvLocalSetup.pull_progress_cb = dem.cli.core_cb.pull_progress_cb
    DevEnvLocalSetup.status_start_cb = dem.cli.core_cb.status_start_cb
    DevEnvLocalSetup.status_stop_cb = dem.cli.core_cb.status_stop_cb

    try:
        dem.cli.main.typer_cli(prog_name=__command__)
    except LookupError as e:
        stderr.print("[red]" + str(e) + "[/]")
    except RegistryError as e:
        stderr.print("[red]" + str(e) + "[/]")
    except docker.errors.DockerException as e:
        stderr.print("[red]" + str(e) + "[/]")

        if "Permission denied" in str(e):
            stdout.print("\nHint: Is your user part of the docker group?")
        elif "invalid reference format" in str(e):
            stdout.print("\nHint: The input repository might not exist in the registry.")
        else:
            stdout.print("\nHint: Probably something is wrong with your Docker Engine installation. Try to reinstall it.")

# Call the main() when run as `python -m`
if __name__ == "__main__":
    main()



"""rename CLI command implementation."""
# dem/cli/command/rename_cmd.py

from dem.core.dev_env_setup import DevEnvLocalSetup
from dem.cli.console import stderr

def execute(dev_env_name_to_rename: str, new_dev_env_name: str) -> None:
    dev_env_local_setup = DevEnvLocalSetup()
    dev_env_to_rename = dev_env_local_setup.get_dev_env_by_name(dev_env_name_to_rename)

    if dev_env_to_rename is not None:
        dev_env_to_rename.name = new_dev_env_name
        dev_env_local_setup.update_json()
    else:
        stderr.print("[red]Error: The input Development Environment does not exist.[/]")


"""The pull command"""
# dem/cli/command/pull_cmd.py

from dem.core import dev_env_setup as dev_env_setup, \
                     registry as registry
from dem.core.tool_images import ToolImages
from dem.cli.console import stdout, stderr

def install_to_dev_env_json(dev_env_local: dev_env_setup.DevEnvLocal, 
                            dev_env_org: dev_env_setup.DevEnvOrg,
                            dev_env_local_setup: dev_env_setup.DevEnvLocalSetup) -> dev_env_setup.DevEnvLocal:
    """Install the Dev Env descriptor to the dev_env.json.
    
    If the dev_env_local is None, the Dev Env is not yet installed locally. Install it by appending 
    the Dev Env descriptor to the dev_env.json file.
    If the Dev Env is already installed, but the descriptor is different, then overwrite the 
    dev_env.json element with the new one.
    If the Dev Env is already installed and both the descriptors are the same, then there is nothing
    to do. 
    Args:
        dev_env_local -- local Dev Env instance (none if not yet installed)
        dev_env_org -- organization's Dev Env instance
        dev_env_local_setup -- the local Dev Env setup
    Return with the dev_env_local instance.
    """
    if dev_env_local is None:
        # If not available, install it.
        dev_env_local = dev_env_setup.DevEnvLocal(dev_env_org=dev_env_org)
        dev_env_local_setup.dev_envs.append(dev_env_local)
        dev_env_local_setup.update_json()
    elif dev_env_local.tools != dev_env_org.tools:
        # If already installed, but different, then overwrite it.
        dev_env_local.tools = dev_env_org.tools
        dev_env_local_setup.update_json()

    return dev_env_local

def execute(dev_env_name: str) -> None:
    # Get the organization's Dev Env if available.
    dev_env_org_setup = dev_env_setup.DevEnvOrgSetup()
    dev_env_org = dev_env_org_setup.get_dev_env_by_name(dev_env_name)
    if dev_env_org is None:
        stderr.print("[red]Error: The input Development Environment is not available for the organization.[/]")
        return

    dev_env_local_setup = dev_env_setup.DevEnvLocalSetup()
    dev_env_local = dev_env_org.get_local_instance(dev_env_local_setup)
    dev_env_local = install_to_dev_env_json(dev_env_local, dev_env_org, dev_env_local_setup)

    # The local Dev Env setup contains the DevEnvOrg to install. Check the images' statuses
    dev_env_local.check_image_availability(dev_env_local_setup.tool_images)
    dev_env_local_setup.pull_images(dev_env_local.tools)
    # Check image availability again.
    image_statuses = dev_env_local.check_image_availability(dev_env_local_setup.tool_images, 
                                                            update_tool_images=True)

    if image_statuses.count(ToolImages.LOCAL_AND_REGISTRY) == len(image_statuses):
        stdout.print("The [yellow]" + dev_env_local.name + "[/] Development Environment is ready!")
    else:
        stderr.print("The installation failed.")


"""info CLI command implementation."""
# dem/cli/command/info_cmd.py

from dem.core.tool_images import ToolImages
from dem.core.dev_env_setup import DevEnvLocal, DevEnvOrg, DevEnvLocalSetup, DevEnvOrgSetup
from dem.cli.console import stdout, stderr
from rich.table import Table

image_status_messages = {
    ToolImages.NOT_AVAILABLE: "[red]Error: Image is not available.[/]",
    ToolImages.LOCAL_ONLY: "Image is available locally.",
    ToolImages.REGISTRY_ONLY: "Image is available in the registry.",
    ToolImages.LOCAL_AND_REGISTRY: "Image is available locally and in the registry.",
}

def print_info(dev_env: (DevEnvLocal | DevEnvOrg)) -> None:
    tool_info_table = Table()
    tool_info_table.add_column("Type")
    tool_info_table.add_column("Image")
    tool_info_table.add_column("Status")
    for tool in dev_env.tools:
        tool_info_table.add_row(tool["type"], tool["image_name"] + ':' + tool["image_version"],
                                image_status_messages[tool["image_status"]])
    stdout.print(tool_info_table)

def execute(arg_dev_env_name: str) -> None:
    dev_env_setup = DevEnvLocalSetup()
    dev_env = dev_env_setup.get_dev_env_by_name(arg_dev_env_name)

    if dev_env is None:
        dev_env_setup = DevEnvOrgSetup()
        dev_env = dev_env_setup.get_dev_env_by_name(arg_dev_env_name)

    if dev_env is None:
        stderr.print("[red]Error: Unknown Development Environment: " + arg_dev_env_name + "[/]")
        return
    else:
        dev_env.check_image_availability(dev_env_setup.tool_images)
        print_info(dev_env)